import pandas as pd
from .db import Db
from .charts import Plot
from .data import Df
from .report import Report
from .errors import Errors
from .messages import Messages

__version__ = "0.4.4"


class DataSwim(Db, Df, Plot, Report, Errors, Messages):
    """
    Main class
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        global __version__
        self.version = __version__
        self.df = df
        self.db = db
        self.x = None
        self.y = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = {}
        self.engine = "bokeh"
        self.label = None
        self.backup_df = None
        self.autoprint = True
        self.errors_handling = "exceptions"
        self.notebook = False
        self.header = self._header
        self.footer = self._footer
        self.reports = []
        self.report_engines = [self.engine]
        self.start_time = None
        self.influx_cli = None
        self.datapath = None
        self.report_path = None
        self.quiet = False

    def __repr__(self):
        num = 0
        if self.df is not None:
            num = len(self.df.index)
        msg = "<DataSwim object | " + str(num) + " rows>"
        return msg

    def new_(self, df=None, db=None, quiet=False):
        """
        Returns a new DataSwim instance from a dataframe
        """
        try:
            ds2 = DataSwim(df, db)
        except Exception as e:
            self.err(e, self.new_, "Can not set new instance")
        if self.autoprint is True and quiet is False:
            self.ok("A new instance was created")
        return ds2

    def set_df(self, data, **args):
        """
        Set a dataframe and an instance
        """
        self.df = pd.DataFrame(data, **args)


ds = DataSwim()



# -*- coding: utf-8 -*-

import pandas as pd


class Transform():
    """
    Class to transform data
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df
        self.db = db

    def keep_(self, *fields):
        """
        Limit a dataframe to some columns
        """
        try:
            return self.clone_(self.df[list(fields)].copy())
        except Exception as e:
            self.err(e, self.keep_, "Can not remove colums")
        if self.autoprint is True:
            self.ok("Columns", " ,".join(fields), "kept")

    def keep(self, *fields):
        """
        Limit a dataframe to some columns
        """
        try:
            self.df = self.df[list(fields)].copy()
        except Exception as e:
            self.err(e, self.keep, "Can not remove colums")
        if self.autoprint is True:
            self.ok("Setting dataframe to columns", " ".join(fields))

    def drop(self, *cols):
        """
        Drops a column from the main dataframe
        """
        index = self.df.columns.values
        for col in cols:
            if col not in index:
                self.warning("Column", col, "not found. Aborting.")
                return
        try:
            for col in cols:
                self.df = self.df.drop(col, axis=1)
        except Exception as e:
            self.err(e, self.drop, "Can not drop column")

    def exclude(self, col, val):
        """
        Delete rows based on value
        """
        try:
            self.df = self.df[self.df[col] != val]
        except Exception as e:
            self.err(e, self.exclude,
                     "Can not exclude rows based on value " + val)

    def rsum(self, time_period="1Min", num_col="num", dateindex=None, index_col="date", fill_col=None):
        """
        Resample, and sum the main dataframe to a time period
        """
        try:
            self.df = self._resample("sum", time_period,
                                     num_col, dateindex, index_col, fill_col)
        except Exception as e:
            self.err(e, self.rsum, "Can not sum data")

    def rsum_(self, time_period="1Min", num_col="num", dateindex=None, index_col="date", fill_col=None):
        """
        Resample, and sum a dataframe to a time period
        """
        try:
            df = self._resample("sum", time_period,
                                num_col, dateindex, index_col, fill_col)
            if num_col is not None:
                self.add(0, num_col)
        except Exception as e:
            self.err(e, self.rsum_, "Can not sum data")
            return
        return self.clone_(df)

    def _resample(self, method, time_period, num_col, dateindex, index_col, fill_col):
        """
        Resample the main dataframe to a time period
        """
        try:
            ds2 = self.clone_()
            if dateindex is not None:
                try:
                    ds2 = ds2.dateindex_(dateindex)
                    dateindex = None
                except Exception as e:
                    self.err(e, self._resample, "Can not process date index")
                    return
            if num_col is not None:
                ds2.add(num_col, 1)
            ds2.df = ds2.df.resample(time_period)
            if method == "sum":
                ds2.df = ds2.df.sum()
            elif method == "mean":
                ds2.df = ds2.df.mean()
            else:
                self.err(self._resample, "Resampling method " +
                         method + " unknown")
                return
            if self.autoprint is True:
                self.ok("Data resampled by", time_period)
            try:
                ds3 = self.transform_(None, index_col,
                                      fill_col, None, df=ds2.df)
            except Exception as e:
                self.err(e, self._resample,
                         "Can not transform data after resampling")
            return ds3.df
        except Exception as e:
            self.err(e, self._resample, "Can not resample data")
            return

    def rmean(self, time_period="1Min", num_col="num", dateindex=None, index_col="date", fill_col=None):
        """
        Resample, and sum the main dataframe to a time period
        """
        try:
            self.df = self._resample("mean", time_period,
                                     num_col, dateindex, index_col, fill_col)
        except Exception as e:
            self.err(e, self.rmean, "Can not mean data")

    def rmean_(self, time_period="1Min", num_col="num", dateindex=None, index_col="date", fill_col=None):
        """
        Resample, and sum a dataframe to a time period
        """
        try:
            df = self._resample("mean", time_period,
                                num_col, dateindex, index_col, fill_col)
        except Exception as e:
            self.err(e, self.rmean_, "Can not mean data")
            return
        return self.clone_(df)

    def sum_(self, column):
        """
        Returns the sum of all values in a column
        """
        try:
            df = self.df[column]
            num = df.sum()
            return num
        except Exception as e:
            self.err(e, self.sum_, "Can not sum data on column " + column)

    def reverse(self):
        """
        Reverses the main dataframe order
        """
        try:
            self.df = self.df.iloc[::-1]
        except Exception as e:
            self.err(e, self.reverse, "Can not reverse the dataframe")

    def sort(self, column):
        """
        Sorts the main dataframe according to the given column
        """
        try:
            self.df = self.df.copy().sort_values(column)
        except Exception as e:
            self.err(e, self.reverse,
                     "Can not sort the dataframe from column " + column)

    def apply(self, function, cols=None, axis=1):
        """
        Apply a function on columns values
        """
        try:
            if cols is None:
                obj = self.df
                self.df = obj.apply(function, axis=axis)
            else:
                obj = self.df[cols]
                self.df[cols] = obj.apply(function, axis=axis)
        except Exception as e:
            self.err(e, self.apply, "Can not apply function")

    def pivot(self, index, columns, values):
        """
        Pivots a dataframe
        """
        try:
            self.df = self._pivot(index, columns, values)
        except Exception as e:
            self.err(e, self._pivot, "Can not pivot table")

    def pivot_(self, index, columns, values):
        """
        Pivots a dataframe
        """
        try:
            return self.clone_(self._pivot(index, columns, values))
        except Exception as e:
            self.err(e, self._pivot, "Can not pivot table")

    def _pivot(self, index, columns, values):
        """
        Pivots a dataframe
        """
        try:
            return self.df.pivot(index=index, columns=columns, values=values)
        except Exception as e:
            self.err(e, self._pivot, "Can not pivot table")

    def concat(self, *dfs):
        """
        Concatenate dataframes from a list and set it to the main dataframe
        """
        try:
            self.df = self._concat(*dfs)
        except Exception as e:
            self.err(e, self.concat, "Can not concatenate data")

    def concat_(self, *dfs):
        """
        Concatenate dataframes from a list and returns a new DataSwim instance
        """
        try:
            ds2 = self.clone_(self._concat(*dfs))
            return ds2
        except Exception as e:
            self.err(e, self.concat, "Can not concatenate data")

    def _concat(self, *dfs):
        """
        Concatenate dataframes from a list and set it to the main dataframe
        """
        try:
            return pd.concat(dfs)
        except Exception as e:
            self.err(e, self.concat, "Can not concatenate data")

    def split_(self, col):
        """
        Split the main dataframe according to column values
        """
        df = self.df.copy()
        try:
            dss = {}
            unique = df[col].unique()
            for key in unique:
                df2 = df.loc[df[col] == key]
                ds2 = self.clone_(df2)
                dss[key] = ds2
            return dss
        except Exception as e:
            self.err(e, self.split_, "Can not split dataframe")

    def add(self, column, value):
        """
        Add a columns with default values
        """
        try:
            df = self.df.copy()
            df[column] = value
            self.df = df
        except Exception as e:
            self.err(e, self.add, "Can not add column")

    def copy_col(self, origin_col, end_col):
        """
        Copy a columns values in another column
        """
        try:
            df = self.df.copy()
            df[end_col] = df[[origin_col]]
            self.df = df
        except Exception as e:
            self.err(e, self.copy_col, "Can not copy column")

    def index_col(self, column="index"):
        """
        Add a column filled from the index to the main dataframe
        """
        try:
            self.df = self._index_col(column)
        except Exception as e:
            self.err(e, self.index_col, "Can not add index column")

    def index_col_(self, column="index"):
        """
        Returns a DatasWim instance with a new column filled from the index
        """
        try:
            df = self._index_col(column)
            return self.clone_(df=df)
        except Exception as e:
            self.err(e, self.index_col_, "Can not add index column")

    def _index_col(self, column):
        """
        Add a column from the index
        """
        df = self.df.copy()
        try:
            df[column] = df.index.values
            if self.autoprint is True:
                self.ok("Column", column, "added from index")
            return df
        except Exception as e:
            self.err(e)

    def drop_rows(self, *rows):
        """
        Drops some rows from the main dataframe
        """
        try:
            self.df = self._drop_rows(rows)
        except Exception as e:
            self.err(e, self.drop_rows, "Can not drop rows")
            return

    def _drop_rows(self, rows):
        """
        Drops some rows from the main dataframe
        """
        try:
            df = self.df.copy()
            rows = list(rows)
            df = df.drop(rows)
            return df
        except Exception as e:
            self.err(e, self._drop_rows, "Can not drop rows")
            return
        if self.autoprint is True:
            self.ok("Dropped rows", ", ".join(rows))

    def append(self, vals, index=None):
        """
        Append a row to the main dataframe
        """
        try:
            self.set(self._append(vals, index))
        except Exception as e:
            self.err(e, self.append, "Can not append row")
            return

    def append_(self, vals, index=None):
        """
        Returns a dataswim instance with one more row
        """
        try:
            return self.duplicate_(self._append(vals, index))
        except Exception as e:
            self.err(e, self.append_, "Can not append row")
            return

    def _append(self, vals, index):
        """
        Returns a dataframe with a new row
        """
        try:
            df = self.df.copy()
            if index is None:
                index = len(df)
            df.loc[index] = vals
            return df
        except Exception as e:
            self.err(e, self._append, "Can not append row")
            return
        if self.autoprint is True:
            self.ok("Row added to dataframe")

    def rename(self, source_col, dest_col):
        """
        Renames a column in the main dataframe
        """
        try:
            self.df = self.df.rename(columns={source_col: dest_col})
        except Exception as e:
            self.err(e, self.rename, "Can not rename column")
            return
        if self.autoprint is True:
            self.ok("Column", source_col, "renamed")



from influxdb import DataFrameClient


class InfluxDb():
    """
    A class to handle InfluxDb database
    """

    def __init__(self):
        """
        Initialize
        """
        self.influx_cli = None

    def influx_init(self, url, port, user, pwd, db):
        """
        Initialize an Influxdb database client
        """
        try:
            cli = DataFrameClient(url, port, user, pwd, db)
            self.influx_cli = cli
        except Exception as e:
            self.err(e, self.influx_init,
                     "Can not initialize Influxdb database")

    def influx_to_csv(self, measurement, batch_size=5000):
        """
        Batch export data from an Influxdb measurement to csv
        """
        if self.autoprint is True:
            self.start("Processing data from InfluxDb to csv")
        try:
            q = "select count(*) from " + measurement
            numrows = self.influx_cli.query(q)[measurement].iloc[0, 0]
            self.info("Processing", str(numrows),
                      "datapoints from InfluxDb to csv")
            done = 0
            offset = 1
            while (done + 1) < numrows:
                q = "select * from " + measurement + \
                    " limit " + str(batch_size)
                if done > 0:
                    q += " offset " + str(offset)
                res = self.influx_cli.query(q)
                numres = 0
                try:
                    numres = len(res[measurement])
                except KeyError:
                    return
                df = res[measurement]
                ds2 = self.clone_(df)
                ds2.to_csv("data/" + measurement +
                           ".csv", index=False, mode='a')
                ds2.df = None
                if numres < batch_size:
                    done += numres
                else:
                    done += batch_size
                self.progress("Processed points", offset, "to", done)
                offset += batch_size
                # if (done + 1) == numrows:
                #    break
        except Exception as e:
            self.err(e, self.influx_to_csv,
                     "Can not transfer data from Influxdb to csv")
            return
        if self.autoprint is True:
            self.end("Finished processing data from InfluxDb to csv")



from goerr.colors import colors


class Search():
    """
    Class to handle searches in the dataframe
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df

    def contains_(self, column, value):
        """
        Returns rows that contains a string value in a column
        """
        try:
            df = self.df[self.df[column].str.contains(value) == True]
            return self.clone_(df.copy())
        except KeyError:
            self.err(self.contains_, "Can not find " +
                     colors.bold(column) + " column")
            return
        except Exception as e:
            self.err(e, self.contains_, "Can not select contained data")

    def exact_(self, column, *values):
        """
        Returns rows that has the exact string value in a column
        """
        try:
            df2 = self.df[column].isin(values)
            df = self.df[df2]
            return self.clone_(df)
        except KeyError:
            self.err(self.exact_, "Can not find " +
                     colors.bold(column) + " column")
            return
        except Exception as e:
            self.err(e, self.exact_, "Can not select exact data")



from numpy.core.numeric import nan
import dataset
import pandas as pd
from goerr import err
from .infos import Info
from .select import Select
from .relations import Relation
from .insert import Insert
from .influxdb import InfluxDb


class Db(Info, Select, Insert, Relation, InfluxDb):
    """
    Class for manipulating databases
    """

    def __init__(self, db=None):
        """
        Initialize with an empty db
        """
        self.db = db

    def connect(self, url):
        """
        Connect to the database and set it as main database
        """
        try:
            self.db = dataset.connect(url)
        except Exception as e:
            self.err(e, self.connect, "Can not connect to database")
            return
        if self.db is None:
            self.err("Database" + url + " not found", self.connect)
            return
        if self.autoprint is True:
            self.ok("Db", self.db.url, "connected")

    def load(self, table, dateindex=None, index_col=None, fill_col=None):
        """
        Set the main dataframe from a table's data
        """
        try:
            self.df = self._load(table, dateindex, index_col, fill_col)
        except Exception as e:
            self.err(e, self.load, "Can not load table " + table)

    def load_(self, table, dateindex=None, index_col=None, fill_col=None):
        """
        Returns a DataSwim instance from a table's data
        """
        try:
            df = self._load(table, dateindex, index_col, fill_col)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.load_, "Can not load table " + table)

    def _load(self, table, dateindex, index_col, fill_col):
        """
        Set the main dataframe or return table's data
        """
        self._check_db()
        try:
            df = self.getall(table)
        except Exception as e:
            self.err(e, self._load, "Can not fetch data from table " + table)
            return self.df
        try:
            ds2 = self.transform_(dateindex, index_col, fill_col, df=df)
        except Exception as e:
            self.err(e, self.load, "Can not transform dataframe")
            return self.df
        if self.autoprint is True:
            self.ok("Data loaded from table", table)
        return ds2.df

    def load_django(self, query, dateindex=None, index_col=None, fill_col=None):
        """
        Returns a DataSwim instance from a django orm query
        """
        try:
            self.df = self._load_django(query, dateindex, index_col, fill_col)
        except Exception as e:
            self.err(e, self.load_django, "Can not load data from query")

    def load_django_(self, query, dateindex=None, index_col=None, fill_col=None):
        """
        Returns a DataSwim instance from a django orm query
        """
        try:
            df = self._load_django(query, dateindex, index_col, fill_col)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.load_django_, "Can not load data from query")

    def _load_django(self, query, dateindex, index_col, fill_col):
        """
        Returns a DataSwim instance from a django orm query
        """
        try:
            df = pd.DataFrame(list(query.values()))
        except Exception as e:
            self.err(e)
            return
        try:
            ds2 = self.transform_(dateindex, index_col, fill_col, df=df)
            df = ds2.df
        except Exception as e:
            self.err(e, self.load, "Can not transform dataframe")
            return
        if self.autoprint is True:
            self.ok("Loaded data from django orm query")
        return df

    def csv_to_db(self):
        """
        Batch transfer data from csv to database
        """
        pass

    def _check_db(self):
        """
        Checks the database connection
        """
        if self.db is None:
            self.err(self._check_db, "Database not connected")



# -*- coding: utf-8 -*-

import pandas as pd
from .views import View
from .clean import Clean
from .count import Count
from .select import Select
from .transform import Transform
from .export import Export
from .search import Search


class Df(Select, View, Transform, Clean, Count, Export, Search):
    """
    Class for manipulating dataframes
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df
        self.datapath = None

    def duplicate_(self, df=None, db=None, quiet=False):
        """
        Returns a new DataSwim instance using the previous database connection
        """
        try:
            if db is None:
                db = self.db
            if df is None:
                if self.df is None:
                    self.err(
                        self.duplicate_, "The main dataframe is empty and no dataframe"
                        " was provided: please provide a dataframe as argument")
                    return
                df = self.df.copy()
            ds2 = self.new_(df, db, quiet=True)
            ds2.db = self.db
            ds2.x = self.x
            ds2.y = self.y
            ds2.chart_obj = self.chart_obj
            ds2.chart_opts = self.chart_opts
            ds2.chart_style = self.chart_style
            ds2.label = self.label
            ds2.reports = self.reports
            ds2.report_engines = self.report_engines
            ds2.backup_df = self.backup_df
            ds2.autoprint = self.autoprint
            ds2.errors_handling = self.errors_handling
            ds2.datapath = self.datapath
            ds2.report_path = self.report_path
        except Exception as e:
            self.err(e, self.duplicate_, "Can not duplicate instance")
            return
        if self.autoprint is True and quiet is False:
            self.ok("A duplicated instance was created")
        return ds2

    def clone_(self, df=None, db=None):
        """
        Silently clone the DataSwim instance
        """
        try:
            ds2 = self.duplicate_(df, db, quiet=True)
            return ds2
        except Exception as e:
            self.err(e, self.clone_, "Can not clone instance")

    def set(self, df):
        """
        Set a main dataframe
        """
        try:
            self.df = df.copy()
        except Exception as e:
            self.err(e, self.set, "Can not set the main dataframe")

    def backup(self):
        """
        Backup the main dataframe
        """
        try:
            self.backup_df = self.df.copy()
        except Exception as e:
            self.err(e, self.backup, "Can not backup data")
        if self.autoprint is True:
            self.ok("Dataframe backed up")

    def restore(self):
        """
        Restore the main dataframe
        """
        try:
            self.df = self._restore()
        except Exception as e:
            self.err(e, self.restore, "Can not restore dataframe")

    def restore_(self):
        """
        Returns the restored main dataframe in a DataSwim instance
        """
        try:
            return self.clone_(self._restore())
        except Exception as e:
            self.err(e, self.restore_, "Can not restore dataframe")

    def _restore(self):
        """
        Restore the main dataframe
        """
        if self.backup_df is None:
            self.warning("No dataframe is backed up: nothing restore")
            return
        try:
            return self.backup_df
        except Exception as e:
            self.err(e)
        if self.autoprint is True:
            self.ok("Dataframe is restored")

    def load_csv(self, url, dateindex=None, index_col=None, fill_col=None, **kwargs):
        """
        Initialize the main dataframe from csv data
        """
        try:
            self.df = self._load_csv(
                url, dateindex, index_col, fill_col, **kwargs)
        except Exception as e:
            self.err(e, self.load_csv, "Can not load csv file")

    def load_csv_(self, url, dateindex=None, index_col=None, fill_col=None, **kwargs):
        """
        Returns a DataSwim instance from csv data
        """
        try:
            df = self._load_csv(url, dateindex, index_col, fill_col, **kwargs)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.load_csv_, "Can not load csv file")

    def _load_csv(self, url, dateindex, index_col, fill_col, **kwargs):
        """
        Returns a DataSwim instance from csv data
        """
        if self.autoprint is True:
            self.start("Loading csv...")
        try:
            df = pd.read_csv(url, **kwargs)
            ds2 = self.clone_(df=df)
        except FileNotFoundError as e:
            msg = "File " + url + " not found"
            self.warning(msg)
            return
        except Exception as e:
            self.err(e, self._load_csv, "Can not load csv file")
            return
        try:
            ds2 = self.transform_(dateindex, index_col, fill_col, df=df)
        except Exception as e:
            self.err(e)
            return
        if self.autoprint is True:
            self.end("Finished loading csv")
        return ds2.df



# -*- coding: utf-8 -*-

import pytablewriter


class Export():
    """
    Class to handle data exportation
    """

    def to_html_(self):
        """
        Exports the main dataframe to html
        """
        renderer = pytablewriter.HtmlTableWriter
        data = self._build_export(renderer)
        return data

    def to_json_(self):
        """
        Exports the main dataframe to json
        """
        renderer = pytablewriter.JsonTableWriter
        data = self._build_export(renderer)
        return data

    def to_markdown_(self):
        """
        Exports the main dataframe to markdown
        """
        renderer = pytablewriter.MarkdownTableWriter()
        data = self._build_export(renderer)
        return data

    def to_rst_(self):
        """
        Exports the main dataframe to restructured text
        """
        renderer = pytablewriter.RstGridTableWriter
        data = self._build_export(renderer)
        return data

    def to_javascript_(self, table_name="table"):
        """
        Exports the main dataframe to javascript
        """
        renderer = pytablewriter.JavaScriptTableWriter
        data = self._build_export(renderer, table_name)
        return data

    def to_python_(self, table_name="table"):
        """
        Exports the main dataframe to python code
        """
        renderer = pytablewriter.PythonCodeTableWriter
        data = self._build_export(renderer, table_name)
        return data

    def to_numpy_(self, table_name="table"):
        """
        Exports the main dataframe to a numpy array
        """
        renderer = pytablewriter.NumpyTableWriter
        data = self._build_export(renderer, table_name)
        return data

    def to_excell(self, filepath, title):
        """
        Writes the main dataframe to an Excell file
        """
        writer = pytablewriter.ExcelXlsxTableWriter()
        writer.open(filepath)
        writer.make_worksheet(title)
        writer.write_table_iter()
        writer.close()
        if self.autoprint is True:
            self.ok("File exported to", filepath)

    def to_csv(self, filepath, index=False, **args):
        """
        Saves the main dataframe to a csv file
        """
        if self.datapath is not None:
            if filepath.startswith("/") is False:
                filepath = self.datapath + "/" + filepath
        self.df.to_csv(filepath, encoding='utf-8', index=index, **args)
        if self.autoprint is True:
            self.ok("Data exported to", filepath)

    def _build_export(self, renderer, table_name=None):
        """
        Builds an export with a renderer
        """
        writer = renderer()
        writer.from_dataframe(self.df)
        if table_name is not None:
            writer.table_name = table_name
        data = writer.write_table()
        return data



import seaborn as sns


class Seaborn():
    """
    A class to handle Seaborn charts
    """

    def residual_(self, label=None, style=None, opts=None):
        """
        Returns a Seaborn models residuals chart
        """
        opts, style = self._get_opts(opts, style)
        self._check_defaults(x_only=True)
        self._set_seaborn_engine()
        color, _ = self._get_color_size(style)
        try:
            fig = sns.residplot(self.df[self.x], self.df[self.y],
                                lowess=True, color=color)
            fig = self._set_with_height(fig, opts)
            return fig
        except Exception as e:
            self.err(e, self.residual_,
                     "Can not draw models residuals chart")

    def density_(self, label=None, style=None, opts=None):
        """
        Returns a Seaborn density chart
        """
        opts, style = self._get_opts(opts, style)
        self._check_defaults(x_only=True)
        self._set_seaborn_engine()
        try:
            fig = sns.kdeplot(self.df[self.x], self.df[self.y])
            fig = self._set_with_height(fig, opts)
            return fig
        except Exception as e:
            self.err(e, self.density_,
                     "Can not draw density chart")

    def distrib_(self, label=None, style=None, opts=None):
        """
        Returns a Seaborn distribution chart
        """
        opts, style = self._get_opts(opts, style)
        self._check_defaults(x_only=True)
        self._set_seaborn_engine()
        color, _ = self._get_color_size(style)
        try:
            fig = sns.distplot(self.df[self.x], color=color)
            fig = self._set_with_height(fig, opts)
            return fig
        except Exception as e:
            self.err(e, self.distrib_,
                     "Can not draw distribution chart")

    def linear_(self, label=None, style=None, opts=None):
        """
        Returns a Seaborn linear regression plot
        """
        opts, style = self._get_opts(opts, style)
        self._check_defaults()
        self._set_seaborn_engine()
        xticks, yticks = self._get_ticks(opts)
        color, size = self._get_color_size(style)
        self.chart_type = "linear"
        try:
            fig = sns.lmplot(self.x, self.y, data=self.df)
            fig = self._set_with_height(fig, opts)
            return fig
        except Exception as e:
            self.err(e, self.linear_,
                     "Can not draw linear regression chart")

    def dlinear_(self, label=None, style=None, opts=None):
        """
        Returns a Seaborn linear regression plot with marginal distribution
        """
        opts, style = self._get_opts(opts, style)
        self._check_defaults()
        self._set_seaborn_engine()
        xticks, yticks = self._get_ticks(opts)
        color, size = self._get_color_size(style)
        self.chart_type = "linear"
        try:
            fig = sns.jointplot(self.x, self.y, data=self.df, kind="reg",
                                xlim=xticks, ylim=yticks, color=color, size=size)
            fig = self._set_with_height(fig, opts)
            return fig
        except Exception as e:
            self.err(e, self.dlinear_,
                     "Can not draw linear regression chart with distribution")

    def _get_seaborn_chart(self, xfield, yfield, chart_type, label, opts=None, style=None, **kwargs):
        """
        Get an Seaborn chart object
        """
        opts, style = self._get_opts(opts, style)
        # params
        opts["xfield"] = xfield
        opts["yfield"] = yfield
        opts["dataobj"] = self.df
        opts["chart_type"] = chart_type
        opts["options"] = options
        sns.set(style="darkgrid", color_codes=True)
        # generate
        try:
            if chart_type == "dlinear":
                chart_obj = self.dlinear_(label, style, opts, options)
            elif chart_type == "linear":
                chart_obj = self.linear_(label, style, opts, options)
            elif chart_type == "distribution":
                chart_obj = self.distrib_(label, style, opts, options)
            elif chart_type == "density":
                chart_obj = self.density_(label, style, opts, options)
            elif chart_type == "residual":
                chart_obj = self.residual_(label, style, opts, options)
            else:
                self.err(self._get_seaborn_chart, "Chart type " +
                         chart_type + " not supported with Seaborn")
                return
        except Exception as e:
            self.err(e, self._get_seaborn_chart,
                     "Can not get Altair chart object")
            return
        return chartobj

    def _get_ticks(self, opts):
        """
        Check if xticks and yticks are set
        """
        opts, _ = self._get_opts(opts, None)
        if not "xticks" in opts:
            if not "xticks" in self.chart_opts:
                self.err(self.dlinear_,
                         "Please set the xticks option for this chart to work")
                return
            else:
                xticks = self.chart_opts["xticks"]
        else:
            xticks = opts["xticks"]
        if not "yticks" in opts:
            if not "yticks" in self.chart_opts:
                self.err(self.dlinear_,
                         "Please set the yticks option for this chart to work")
                return
            else:
                yticks = self.chart_opts["yticks"]
        else:
            yticks = opts["yticks"]
        return xticks, yticks

    def _get_color_size(self, style):
        """
        Get color and size from a style dict
        """
        color = "b"
        if "color" in style:
            color = style["color"]
        size = 7
        if "size" in style:
            size = style["size"]
        return color, size

    def _set_with_height(self, fig, opts):
        """
        Set the width and height of a Matplotlib figure
        """
        h = 6
        if "height" in opts:
            h = opts["height"]
        w = 6
        if "width" in opts:
            w = opts["width"]
        try:
            fig.figure.set_size_inches((w, h))
            return fig
        except:
            try:
                fig.fig.set_size_inches((w, h))
                return fig
            except Exception as e:
                self.err(e, self._set_with_height,
                         "Can not set figure width and height from chart object")

    def _get_opts(self, opts, style):
        """
        Get default options
        """
        o = opts
        if opts is None:
            o = self.chart_opts
        s = style
        if style is None:
            s = self.chart_style
        return o, s

    def _save_seaborn_chart(self, report, folderpath):
        """
        Saves a png image of the seaborn chart
        """
        if folderpath is None:
            if self.imgs_path is None:
                self.err(self._save_seaborn_chart,
                         "Please set a path where save images: ds.imgs_path = '/my/path'")
                return
            path = self.imgs_path
        else:
            path = folderpath
        path = path + "/" + report["slug"] + ".png"
        try:
            if type(report["seaborn_chart"]) == seaborn.axisgrid.JointGrid:
                report["seaborn_chart"].savefig(path)
            else:
                report["seaborn_chart"].figure.savefig(path)
        except Exception as e:
            self.err(e, self._save_seaborn_chart, "Can not save Seaborn chart")
            return
        if self.autoprint is True:
            self.ok("Seaborn chart writen to file")

    def _set_seaborn_engine(self):
        """
        Set the current chart engine to Seaborn
        """
        if self.engine != "seaborn":
            self.engine = "seaborn"
            self.info("Switching to the Seaborn engine to draw this chart")



from datetime import datetime
import pandas as pd


class Select():
    """
    Class to select data
    """

    def __init__(self, df=None):
        """
        Initialize with an empty dataframe
        """
        self.df = df

    def first_(self):
        """
        Returns the first row
        """
        return self._first()

    def _first(self):
        """
        Select the first row
        """
        try:
            val = self.df.iloc[0]
            return val
        except Exception as e:
            self.err(e, self.first, "Can not select first row")

    def limit(self, r=5):
        """
        Limit selection to a range in the main dataframe
        """
        try:
            self.df = self.df[:r]
        except Exception as e:
            self.err(e, self.limit, "Can not limit data")

    def limit_(self, r=5):
        """
        Returns a DataSwim instance with limited selection
        """
        try:
            return self.new(self.df[:r])
        except Exception as e:
            self.err(e)

    def unique_(self, column):
        """
        Returns unique values in a column
        """
        try:
            df = self.df.copy()
            df.drop_duplicates(subset=[column], inplace=True)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.unique_, "Can not select unique data")

    def range(self, **args):
        """
        Limit the data in a time range
        """
        try:
            self.df = self._range(**args)
        except Exception as e:
            self.err(e, self.range, "Can not select range data")
            return

    def _range(self, **args):
        """
        Limit the data in a time range
        """
        try:
            df = self.df[self.df.last_valid_index() -
                         pd.DateOffset(**args):]
            return df
        except Exception as e:
            self.err(e, self._range, "Can not select range data")
            return

    def range_(self, **args):
        """
        Limit the data in a time range
        """
        try:
            ds2 = self.clone_(self._range(**args))
            return ds2
        except Exception as e:
            self.err(e, self.range_, "Can not select range data")

    def nowrange(self, col, interval, unit="D"):
        """
        Set the main dataframe with rows within a date range from now
        """
        try:
            df = self._nowrange(col, interval, unit)
            self.set(df)
        except Exception as e:
            self.err(e, self.nowrange_, "Can not select range data from now")

    def nowrange_(self, col, interval, unit="D"):
        """
        Returns a Dataswim instance with rows within a date range from now
        """
        try:
            df = self._nowrange(col, interval, unit)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.nowrange_, "Can not select range data from now")

    def _nowrange(self, col, interval, unit):
        """
        Returns a dataframe with rows within a date range from now
        """
        try:
            df = self.df.copy()
            df = df[df[col].dt.date > datetime.now().date(
            ) - pd.to_timedelta(interval, unit=unit)]
            return df
        except Exception as e:
            self.err(e, self._nowrange, "Can not select range data from now")

    def daterange(self, datecol, date_start, op, **args):
        """
        Returns rows in a date range
        """
        try:
            self.df = self._daterange(datecol, date_start, op, **args)
        except Exception as e:
            self.err(e, self.daterange, "Can not select date range data")

    def daterange_(self, datecol, date_start, op, **args):
        """
        Returns a DataSwim instance with rows in a date range
        """
        try:
            df = self._daterange(datecol, date_start, op, **args)
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.daterange_, "Can not select date range data")

    def _daterange(self, datecol, date_start, op, **args):
        """
        Returns rows in a date range
        """
        try:
            start_date = pd.Timestamp(date_start)
            self.df[datecol] = pd.to_datetime(self.df[datecol])
            if op == "+":
                end_date = start_date + pd.DateOffset(**args)
            elif op == "-":
                end_date = start_date - pd.DateOffset(**args)
            mask = (self.df[datecol] >= start_date) & (
                self.df[datecol] <= end_date)
            df = self.df.loc[mask]
            return df
        except Exception as e:
            self.err(e, self._daterange, "Can not select date range data")

    def to_records_(self):
        """
        Returns a list of dictionary records from the main dataframe
        """
        try:
            dic = self.df.to_dict(orient="records")
            return dic
        except Exception as e:
            self.err(e, self.to_records_, "Can not create records")

    def nulls_(self, field):
        """
        Return all null rows
        """
        try:
            null_rows = self.df[field][self.df[field].isnull()]
        except Exception as e:
            self.err(e, self.nulls_, "Can not select null rows")
        return null_rows
